/*
**
** Filename: eepinic.h
**
** 0.44 2000/07/25 1st Release
** 0.50 2000/11/30 2nd Release
**                 Support for PCI31L (PCB Version 2400)
**
*/

#ifndef CVT_LINUX_VERSION
#define CVT_LINUX_VERSION(V,P,S)	(((V) * 65536) + ((P) * 256) + (S))
#endif

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

// #if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,92)
// #include <linux/bios32.h>
// #endif
// #if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,2,13)
// #include <linux/init.h>
// #endif
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
#include <asm/spinlock.h>
#endif

#include <stdio.h>
#include <stdlib.h>

#ifndef uint
#define uint unsigned int
#endif
#ifndef u_int
#define u_int unsigned int
#endif
#ifndef ushort
#define ushort unsigned short
#endif
#ifndef u_short
#define u_short unsigned short
#endif
#ifndef uchar
#define uchar unsigned char
#endif
#ifndef u_char
#define u_char unsigned char
#endif
#ifndef ulong
#define ulong unsigned long
#endif

#ifndef PVOID
#define PVOID void *
#endif
#ifndef UBYTE
#define UBYTE unsigned char
#endif
#ifndef UCHAR
#define UCHAR unsigned char
#endif
#ifndef UWORD
#define UWORD unsigned short
#endif
#ifndef USHORT
#define USHORT unsigned short
#endif
#ifndef ULONG
#define ULONG unsigned long
#endif

#ifdef ALPHA
#define U32	unsigned int
#else
#define U32	unsigned long
#endif

#ifndef NULL
#define NULL	0	/* zero */
#endif

#ifndef TRUE
#define TRUE	(1)	/* boolean true */
#endif
#ifndef FALSE
#define FALSE	(0)	/* boolean false */
#endif

#define TOTAL_SG_ENTRY	32

/***********************************************************************
		Scatter-Gather Element Structure
************************************************************************/
typedef struct ORC_SG_Struc {
	U32 SG_Ptr;		/* Data Pointer */
	U32 SG_Len;		/* Data Length */
} ORC_SG;

#define MAX_CHANNELS	2

/********************************************************/
/*      Orchid Configuration Register Set               */
/********************************************************/
#define ORC_PVID	0x00	/* Vendor ID                      */
#define ORC_VENDOR_ID	0x1195	/* Orchid vendor ID               */
#define ORC_PDID        0x02	/* Device ID                    */
#define ORC_DEVICE_ID	0x0853	/* Orchid device ID               */
#define ORC_COMMAND	0x04	/* Command                        */
#define BUSMS		0x04	/* BUS MASTER Enable              */
#define IOSPA		0x01	/* IO Space Enable                */
#define ORC_STATUS	0x06	/* Status register                */
#define ORC_REVISION	0x08	/* Revision number                */
#define ORC_BASE	0x10	/* Base address                   */
#define ORC_BIOS	0x50	/* Expansion ROM base address     */
#define ORC_INT_NUM	0x3C	/* Interrupt line         */
#define ORC_INT_PIN	0x3D	/* Interrupt pin          */

/********************************************************/
/*      Orchid Host Command Set                         */
/********************************************************/
#define ORC_CMD_NOP		0x00	/* Host command - NOP             */
#define ORC_CMD_VERSION		0x01	/* Host command - Get F/W version */
#define ORC_CMD_ECHO		0x02	/* Host command - ECHO            */
#define ORC_CMD_SET_NVM		0x03	/* Host command - Set NVRAM       */
#define ORC_CMD_GET_NVM		0x04	/* Host command - Get NVRAM       */
#define ORC_CMD_GET_BUS_STATUS	0x05	/* Host command - Get SCSI bus status */
#define ORC_CMD_ABORT_SCB	0x06	/* Host command - Abort SCB       */
#define ORC_CMD_ISSUE_SCB	0x07	/* Host command - Issue SCB       */
/********************************************************/
/*              Orchid Register Set                     */
/********************************************************/
#define ORC_GINTS	0xA0	/* Global Interrupt Status        */
#define QINT		0x04	/* Reply Queue Interrupt  */
#define ORC_GIMSK	0xA1	/* Global Interrupt MASK  */
#define MQINT		0x04	/* Mask Reply Queue Interrupt     */
#define	ORC_GCFG	0xA2	/* Global Configure               */
#define EEPRG		0x01	/* Enable EEPROM programming */
#define	ORC_GSTAT	0xA3	/* Global status          */
#define WIDEBUS		0x10	/* Wide SCSI Devices connected    */
#define ORC_HDATA	0xA4	/* Host Data                      */
#define ORC_HCTRL	0xA5	/* Host Control                   */
#define SCSIRST		0x80	/* SCSI bus reset         */
#define HDO			0x40	/* Host data out          */
#define HOSTSTOP		0x02	/* Host stop RISC engine  */
#define DEVRST		0x01	/* Device reset                   */
#define ORC_HSTUS	0xA6	/* Host Status                    */
#define HDI			0x02	/* Host data in                   */
#define RREADY		0x01	/* RISC engine is ready to receive */
#define	ORC_NVRAM	0xA7	/* Nvram port address             */
#define SE2CS		0x008
#define SE2CLK		0x004
#define SE2DO		0x002
#define SE2DI		0x001
#define ORC_PQUEUE	0xA8	/* Posting queue FIFO             */
#define ORC_PQCNT	0xA9	/* Posting queue FIFO Cnt */
#define ORC_RQUEUE	0xAA	/* Reply queue FIFO               */
#define ORC_RQUEUECNT	0xAB	/* Reply queue FIFO Cnt           */
#define	ORC_FWBASEADR	0xAC	/* Firmware base address  */

#define	ORC_EBIOSADR0 0xB0	/* External Bios address */
#define	ORC_EBIOSADR1 0xB1	/* External Bios address */
#define	ORC_EBIOSADR2 0xB2	/* External Bios address */
#define	ORC_EBIOSDATA 0xB3	/* External Bios address */

#define	ORC_SCBSIZE	0xB7	/* SCB size register              */
#define	ORC_SCBBASE0	0xB8	/* SCB base address 0             */
#define	ORC_SCBBASE1	0xBC	/* SCB base address 1             */

#define	ORC_RISCCTL	0xE0	/* RISC Control                   */
#define PRGMRST		0x002
#define DOWNLOAD		0x001
#define	ORC_PRGMCTR0	0xE2	/* RISC program counter           */
#define	ORC_PRGMCTR1	0xE3	/* RISC program counter           */
#define	ORC_RISCRAM	0xEC	/* RISC RAM data port 4 bytes     */

typedef struct orc_extended_scb {	/* Extended SCB                 */
	ORC_SG ESCB_SGList[TOTAL_SG_ENTRY];	/*0 Start of SG list              */
	unsigned char *SCB_Srb;	/*50 SRB Pointer */
//         Scsi_Cmnd    *SCB_Srb;       /*50 SRB Pointer */
} ESCB;


/***********************************************************************
		Target Device Control Structure
**********************************************************************/

typedef struct ORC_Tar_Ctrl_Struc {
	UBYTE TCS_DrvDASD;	/* 6 */
	UBYTE TCS_DrvSCSI;	/* 7 */
	UBYTE TCS_DrvHead;	/* 8 */
	UWORD TCS_DrvFlags;	/* 4 */
	UBYTE TCS_DrvSector;	/* 7 */
} ORC_TCS, *PORC_TCS;

/***********************************************************************
              Host Adapter Control Structure
************************************************************************/
typedef struct ORC_Ha_Ctrl_Struc {
	USHORT HCS_Base;	/* 00 */
	UBYTE HCS_Index;	/* 02 */
	UBYTE HCS_Intr;		/* 04 */
	UBYTE HCS_SCSI_ID;	/* 06    H/A SCSI ID */
	UBYTE HCS_BIOS;		/* 07    BIOS configuration */

	UBYTE HCS_Flags;	/* 0B */
	UBYTE HCS_HAConfig1;	/* 1B    SCSI0MAXTags */
	UBYTE HCS_MaxTar;	/* 1B    SCSI0MAXTags */

	USHORT HCS_Units;	/* Number of units this adapter  */
	USHORT HCS_AFlags;	/* Adapter info. defined flags   */
	ULONG HCS_Timeout;	/* Adapter timeout value   */
	PVOID HCS_virScbArray;	/* 28 Virtual Pointer to SCB array     */
	U32 HCS_physScbArray;	/* Scb Physical address */
	PVOID HCS_virEscbArray;	/* Virtual pointer to ESCB Scatter list */
	U32 HCS_physEscbArray;	/* scatter list Physical address */
	UBYTE TargetFlag[16];	/* 30  target configuration, TCF_EN_TAG */
	UBYTE MaximumTags[16];	/* 40  ORC_MAX_SCBS */
	UBYTE ActiveTags[16][16];	/* 50 */
	ORC_TCS HCS_Tcs[16];	/* 28 */
	U32 BitAllocFlag[MAX_CHANNELS][8];	/* Max STB is 256, So 256/32 */
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
	spinlock_t BitAllocFlagLock;
#endif
	ULONG pSRB_head;
	ULONG pSRB_tail;
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
	spinlock_t pSRB_lock;
#endif
} ORC_HCS;

typedef struct _NVRAM {
/*----------header for CardBus---------------*/
	UCHAR CB_SubVendorID0;	/* 00 - Sub Vendor ID           */
	UCHAR CB_SubVendorID1;	/* 00 - Sub Vendor ID           */
	UCHAR CB_SubSysID0;	/* 02 - Sub System ID           */
	UCHAR CB_SubSysID1;	/* 02 - Sub System ID           */
	UCHAR CB_SubClass;	/* 04 - Sub Class               */
	UCHAR CB_VendorID0;	/* 05 - Vendor ID               */
	UCHAR CB_VendorID1;	/* 05 - Vendor ID               */
	UCHAR CB_DeviceID0;	/* 07 - Device ID               */
	UCHAR CB_DeviceID1;	/* 07 - Device ID               */
	UCHAR CB_CISPointer[4];	/* 08 - CIS Pointer             */
	UCHAR CB_Padding[192-13];/* 0C - Padding                */
/*----------header ---------------*/
	UCHAR SubVendorID0;	/* 00 - Sub Vendor ID           */
	UCHAR SubVendorID1;	/* 00 - Sub Vendor ID           */
	UCHAR SubSysID0;	/* 02 - Sub System ID           */
	UCHAR SubSysID1;	/* 02 - Sub System ID           */
	UCHAR SubClass;		/* 04 - Sub Class               */
	UCHAR VendorID0;	/* 05 - Vendor ID               */
	UCHAR VendorID1;	/* 05 - Vendor ID               */
	UCHAR DeviceID0;	/* 07 - Device ID               */
	UCHAR DeviceID1;	/* 07 - Device ID               */
	UCHAR Reserved0[2];	/* 09 - Reserved                */
	UCHAR Revision;		/* 0B - Revision of data structure */
	/* ----Host Adapter Structure ---- */
	UCHAR NumOfCh;		/* 0C - Number of SCSI channel  */
	UCHAR BIOSConfig1;	/* 0D - BIOS configuration 1    */
	UCHAR BIOSConfig2;	/* 0E - BIOS boot channel&target ID */
	UCHAR BIOSConfig3;	/* 0F - BIOS configuration 3    */
	/* ----SCSI channel Structure ---- */
	/* from "CTRL-I SCSI Host Adapter SetUp menu "  */
	UCHAR SCSI0Id;		/* 10 - Channel 0 SCSI ID       */
	UCHAR SCSI0Config;	/* 11 - Channel 0 SCSI configuration */
	UCHAR SCSI0MaxTags;	/* 12 - Channel 0 Maximum tags  */
	UCHAR SCSI0ResetTime;	/* 13 - Channel 0 Reset recovering time */
	UCHAR SCSI0NumOfTarg;	/* 14 - Channel 0 # of Target */
	UCHAR SCSI0ReqSenseDelay;	/* 15 - Channel 0 RequestSense delay time */

	/* ----SCSI target Structure ----  */
	/* from "CTRL-I SCSI device SetUp menu "                        */
	UCHAR Target00Config;	/* 16 - Channel 0 Target 0 config */
	UCHAR Target01Config;	/* 17 - Channel 0 Target 1 config */
	UCHAR Target02Config;	/* 18 - Channel 0 Target 2 config */
	UCHAR Target03Config;	/* 19 - Channel 0 Target 3 config */
	UCHAR Target04Config;	/* 1A - Channel 0 Target 4 config */
	UCHAR Target05Config;	/* 1B - Channel 0 Target 5 config */
	UCHAR Target06Config;	/* 1C - Channel 0 Target 6 config */
	UCHAR Target07Config;	/* 1D - Channel 0 Target 7 config */
	UCHAR Target08Config;	/* 1E - Channel 0 Target 8 config */
	UCHAR Target09Config;	/* 1F - Channel 0 Target 9 config */
	UCHAR Target0AConfig;	/* 20 - Channel 0 Target A config */
	UCHAR Target0BConfig;	/* 21 - Channel 0 Target B config */
	UCHAR Target0CConfig;	/* 22 - Channel 0 Target C config */
	UCHAR Target0DConfig;	/* 23 - Channel 0 Target D config */
	UCHAR Target0EConfig;	/* 24 - Channel 0 Target E config */
	UCHAR Target0FConfig;	/* 25 - Channel 0 Target F config */

	UCHAR SCSI1Id;		/* 26 - Channel 1 SCSI ID       */
	UCHAR SCSI1Config;	/* 27 - Channel 1 SCSI configuration */
	UCHAR SCSI1MaxTags;	/* 28 - Channel 1 Maximum tags  */
	UCHAR SCSI1ResetTime;	/* 29 - Channel 1 Reset recovering time */
	UCHAR SCSI1NumOfTarg;	/* 2A - Channel 1 # of Target */
	UCHAR SCSI1ReqSenseDelay;	/* 2B - Channel 1 RequestSense delay time */

	/* ----SCSI target Structure ----  */
	/* from "CTRL-I SCSI device SetUp menu "                                          */
	UCHAR Target10Config;	/* 2C - Channel 1 Target 0 config */
	UCHAR Target11Config;	/* 2D - Channel 1 Target 1 config */
	UCHAR Target12Config;	/* 2E - Channel 1 Target 2 config */
	UCHAR Target13Config;	/* 2F - Channel 1 Target 3 config */
	UCHAR Target14Config;	/* 30 - Channel 1 Target 4 config */
	UCHAR Target15Config;	/* 31 - Channel 1 Target 5 config */
	UCHAR Target16Config;	/* 32 - Channel 1 Target 6 config */
	UCHAR Target17Config;	/* 33 - Channel 1 Target 7 config */
	UCHAR Target18Config;	/* 34 - Channel 1 Target 8 config */
	UCHAR Target19Config;	/* 35 - Channel 1 Target 9 config */
	UCHAR Target1AConfig;	/* 36 - Channel 1 Target A config */
	UCHAR Target1BConfig;	/* 37 - Channel 1 Target B config */
	UCHAR Target1CConfig;	/* 38 - Channel 1 Target C config */
	UCHAR Target1DConfig;	/* 39 - Channel 1 Target D config */
	UCHAR Target1EConfig;	/* 3A - Channel 1 Target E config */
	UCHAR Target1FConfig;	/* 3B - Channel 1 Target F config */
	USHORT PcbVersion;	/* 3C - Reserved                */
	UCHAR reserved[1];	/* 3E - Reserved                */
	/* ---------- CheckSum ----------       */
	UCHAR CheckSum;		/* 3F - Checksum of NVRam       */
} NVRAM, *PNVRAM;

#define NVRAM_SIZE	sizeof(NVRAM)
#define NVRAM_SET	0xC0

/* Bios Configuration for nvram->BIOSConfig1                            */
#define NBC_BIOSENABLE	0x01	/* BIOS enable                    */
#define NBC_CDROM	0x02	/* Support bootable CDROM */
#define NBC_REMOVABLE	0x04	/* Support removable drive        */

/* Bios Configuration for nvram->BIOSConfig2                            */
#define	NBB_TARGET_MASK	0x0F	/* Boot SCSI target ID number     */
#define	NBB_CHANL_MASK	0xF0	/* Boot SCSI channel number       */

/* Bit definition for nvram->SCSIConfig                                 */
#define NCC_BUSRESET	0x01	/* Reset SCSI bus at power up     */
#define NCC_PARITYCHK	0x02	/* SCSI parity enable             */
#define NCC_LVDS	0x10	/* Enable LVDS                    */
#define NCC_ACTTERM1	0x20	/* Enable active terminator 1     */
#define NCC_ACTTERM2	0x40	/* Enable active terminator 2     */
#define NCC_AUTOTERM	0x80	/* Enable auto termination        */

/* Bit definition for nvram->TargetxConfig                              */
#define	NTC_PERIOD	0x07	/* Maximum Sync. Speed            */
#define NTC_1GIGA	0x08	/* 255 head / 63 sectors (64/32) */
#define NTC_NO_SYNC	0x10	/* NO SYNC. NEGO          */
#define NTC_NO_WIDESYNC	0x20	/* NO WIDE SYNC. NEGO             */
#define	NTC_DISC_ENABLE	0x40	/* Enable SCSI disconnect */
#define NTC_SPINUP	0x80	/* Start disk drive               */

/* Default NVRam values                                                 */
#define NBC_DEFAULT	(NBC_ENABLE)
#define NCC_DEFAULT	(NCC_BUSRESET | NCC_AUTOTERM | NCC_PARITYCHK)
#define NCC_MAX_TAGS	0x20	/* Maximum tags per target        */
#define	NCC_RESET_TIME	0x0A	/* SCSI RESET recovering time     */
#define NTC_DEFAULT	(NTC_1GIGA | NTC_NO_WIDESYNC | NTC_DISC_ENABLE)

#define ORC_RD(x,y)             (UCHAR)(inb(  (int)((ULONG)((ULONG)x+(UCHAR)y)) ))
#define ORC_RDLONG(x,y)         (long)(inl((int)((ULONG)((ULONG)x+(UCHAR)y)) ))

#define ORC_WR(     adr,data)   outb( (UCHAR)(data), (int)(adr))
#define ORC_WRSHORT(adr,data)   outw( (UWORD)(data), (int)(adr))
#define ORC_WRLONG( adr,data)   outl( (ULONG)(data), (int)(adr))

