
/*======================================================================

    PCI31/33, CB32 scsi configuration

    inicconf.c
        0.44 2000/07/27 1st Release
        0.50 2000/11/29 2nd Release
                        Support for PCI31L (PCB Version 2400)
        0.51 2001/03/05 for Redhat 7.0

    The contents of this file are subject to the Mozilla Public
    License Version 1.1 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a copy of
    the License at http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS
    IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
    implied. See the License for the specific language governing
    rights and limitations under the License.

    The initial developer of the original code is RATOC Systems, Inc.
    <www.rexpccard.co.jp>.  Portions created by RATOC Systems, Inc.
    are Copyright (C) 2000 RATOC Systems,Inc.  All Rights Reserved.

    Alternatively, the contents of this file may be used under the
    terms of the GNU Public License version 2 (the "GPL"), in which
    case the provisions of the GPL are applicable instead of the
    above.  If you wish to allow the use of your version of this file
    only under the terms of the GPL and not to allow others to use
    your version of this file under the MPL, indicate your decision
    by deleting the provisions above and replace them with the notice
    and other provisions required by the GPL.  If you do not delete
    the provisions above, a recipient may use your version of this
    file under either the MPL or the GPL.
    
======================================================================*/

#include "eepinic.h"

#include <errno.h>

#ifdef __GLIBC__
#include <sys/io.h>
#else
#include <asm/io.h>
#endif

#define _GNU_SOURCE
#include <getopt.h>

// #include <linux/delay.h>
#include <linux/pci.h>

// extern unsigned long volatile jiffies;

static char *this_program = "PCI31/33, CB32 scsi configuration utility (%s): ";
static char *this_version = "Beta Version 0.50 2000/11/29";

static int dbgf = 0;	/* debug flag  0:normal  1,2,3...:debug */
static int verbose = 0;
static int hex_show = 0;
static int pcbver_set = 0;	/* flag of Set PCB Version */

NVRAM nvram, *nvramp = &nvram;

static UCHAR dftNvRam[256] =
{
/*----------for CardBus-------------*/
    0x95,		/* 0x00: Sub System Vendor ID 0 */
    0x11,		/* 0x01: Sub System Vendor ID 1 */
    0x51,		/* 0x02: Sub System ID 0  851:CB32 853:PCI31/33 */
    0x08,		/* 0x03: Sub System ID 1        */
    0x00,		/* 0x04: SubClass               */
    0x95,		/* 0x05: Vendor ID 0            */
    0x11,		/* 0x06: Vendor ID 1            */
    0x51,		/* 0x07: Device ID 0            */
    0x08,		/* 0x08: Device ID 1            */
    0x3F,0x00,0x00,0x00,	/* 0x09-0x0C:           */
    0xFF,0xFF,0xFF,		/* 0x0D-0x0F: Reserved  */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x10-0x17 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x18-0x1F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x20-0x27 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x28-0x2F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x30-0x37 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x38-0x3F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x40-0x47 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x48-0x4F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x50-0x57 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x58-0x5F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x60-0x67 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x68-0x6F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x70-0x77 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x78-0x7F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x80-0x87 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x88-0x8F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x90-0x97 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x98-0x9F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0xA0-0xA7 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0xA8-0xAF */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0xB0-0xA7 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0xB8-0xBF */
/*----------header -------------*/
	0x95,		/* 0x00: Sub System Vendor ID 0 */
	0x11,		/* 0x01: Sub System Vendor ID 1 */
	0x51,		/* 0x02: Sub System ID 0  851:CB32 853:PCI31/33 */
	0x08,		/* 0x03: Sub System ID 1        */
	0x00,		/* 0x04: SubClass               */
	0x95,		/* 0x05: Vendor ID 0            */
	0x11,		/* 0x06: Vendor ID 1            */
	0x51,		/* 0x07: Device ID 0            */
	0x08,		/* 0x08: Device ID 1            */
	0x00,		/* 0x09: Reserved               */
	0x00,		/* 0x0A: Reserved               */
	0x01,		/* 0x0B: Revision of Data Structure     */
		/* -- Host Adapter Structure --- */
	0x01,		/* 0x0C: Number Of SCSI Channel */
	0x01,		/* 0x0D: BIOS Configuration 1   */
	0x00,		/* 0x0E: BIOS Configuration 2   */
	0x00,		/* 0x0F: BIOS Configuration 3   */
		/* --- SCSI Channel 0 Configuration --- */
	0x07,		/* 0x10: H/A ID                 */
	0x83,		/* 0x11: Channel Configuration  */
	0x20,		/* 0x12: MAX TAG per target     */
	0x0A,		/* 0x13: SCSI Reset Recovering time     */
	0x10,		/* 0x14: # of target */
	0x05,		/* 0x15: RequestSense delay time (3us x n )*/
		/* SCSI Channel 0 Target Configuration  */
			/* 0x16-0x25                    */
				/* bit7 Start-Disk-Drive	*/
				/* bit6 Enable SCSI Disconnection */
				/* bit5 No WDTR Nego */
				/* bit4 Sync Nego */
				/* bit3 255 head/63 sectors */
				/* bit2-0 20MB/S (if Wide, x 2) */
	0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9,
	0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9,
		/* --- SCSI Channel 1 Configuration --- */
	0x07,		/* 0x26: H/A ID                 */
	0x83,		/* 0x27: Channel Configuration  */
	0x20,		/* 0x28: MAX TAG per target     */
	0x0A,		/* 0x29: SCSI Reset Recovering time     */
	0x10,		/* 0x2A: # of target */
	0x05,		/* 0x2B: RequestSense delay time (3us x n )*/
		/* SCSI Channel 1 Target Configuration  */
			/* 0x2C-0x3B                    */
	0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9,
	0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9, 0xE9,
	0x00,		/* 0x3C: PCB Version LowByte */
	0x23,		/* 0x3D: PCB Version HighByte*/
			/*	 PCI33/31:23 PCI31LP:24 */
	0x00,		/* 0x3E: Reserved               */
	0x00		/* 0x3F: Checksum               */
};

static int disc_chg = 0;	/* 0: no change  1:change */
static int sync_chg = 0;	/* 0: no change  1:change */
static int wide_chg = 0;	/* 0: no change  1:change */

static uchar disc_val[16]
	 = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	     0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
		/* 0:disconnect disable  1:enable  0xFF:no set */
static uchar sync_val[16]
	 = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	     0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
		/* 0:Async  5:5MB  10:10MB  20:20MB  0xFF:no set */
static uchar wide_val[16]
	 = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	     0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
		/* 0:wide nego disable  1:enable  0xFF:no set */

static ushort NumTarg = 16;	/* PCI31/33:16      PCI31LP:8 */
static ushort pcb_ver = 0x2300;	/* PCI33/31:0x2300  PCI31LP:0x2400 */

static ORC_HCS hcs;

static uchar sync_org[16], disc_org[16], stdr_org[16], wide_org[16];
static ushort pcb_org;
		/* Do not edit value */

static int ReadDelay = 1;
static int WriteDelay = 20;

#define DvcSleepMilliSecond(n)	usleep(n)
#define waitForPause(n)	usleep(n*100)

/*====================================================================*/
static uchar
waitChipReady(ORC_HCS * hcsp)
/*
Function Name	:waitChipReady
Description	:Wait Inic-850 ready
Argument	:ORC_HCS *hcsp;
			Inic-850 Adapter info.
Result		:TRUE:success   FALSE:time-out
=====================================================================*/
{
    int i;

    for (i = 0; i < 2000; i++) {	/* Wait 1 second for report timeout     */
	/* Wait HOSTSTOP set */
	if (ORC_RD(hcsp->HCS_Base, ORC_HCTRL) & HOSTSTOP)
		return (TRUE);
	waitForPause(5);	/* wait 500ms before try again  */
    }
    return (FALSE);
}

/*====================================================================*/
static uchar
waitHDOoff(ORC_HCS * hcsp)
/*
Function Name	:waitHDOff
Description	:Wait HD0
Argument	:ORC_HCS *hcsp;
			Inic-850 Adapter info.
Result		:TRUE:success   FALSE:time-out
=====================================================================*/
{
    int i;

	/* Wait 1 second for report timeout     */
    for (i = 0; i < 2000; i++) {
	/* Wait HDO off */
	if (!(ORC_RD(hcsp->HCS_Base, ORC_HCTRL) & HDO))
	    return (TRUE);
	waitForPause(5);	/* wait 500ms before try again  */
    }
    return (FALSE);
}

/*====================================================================*/
static uchar
waitHDIset(ORC_HCS * hcsp, UCHAR *pData)
/*
Function Name	:waitHDIset
Description	:Wait HDI
Argument	:ORC_HCS *hcsp;
			Inic-850 Adapter info.
		:UCHAR *pData
Result		:TRUE:success   FALSE:time-out
======================================================================*/
{
    int i;

	/* Wait 1 second for report timeout     */
    for (i = 0; i < 2000; i++) {
	/* Wait HDO off */
	if ((*pData = ORC_RD(hcsp->HCS_Base, ORC_HSTUS)) & HDI)
		return (TRUE);	/* Wait HDI set */
	waitForPause(5);	/* wait 500ms before try again  */
    }
    return (FALSE);
}

/*====================================================================*/
static uchar
set_NVRAM(ORC_HCS * hcsp, UCHAR address, UCHAR value)
/*
Function Name	:set_NVRAM
Description	: 
Argument	:ORC_HCS *hcsp;
			Inic-850 Adapter info.
		:UCHAR address
		:UCHAR value
======================================================================*/
{
	/* Write command */
    ORC_WR(hcsp->HCS_Base + ORC_HDATA, ORC_CMD_SET_NVM);
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    ORC_WR(hcsp->HCS_Base + ORC_HDATA, address);	/* Write address */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    ORC_WR(hcsp->HCS_Base + ORC_HDATA, value);	/* Write value  */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    return (TRUE);
}

/*====================================================================*/
static uchar
get_NVRAM(ORC_HCS * hcsp, unsigned char address, unsigned char *pDataIn)
{
    unsigned char bData;

	/* Write command */
    ORC_WR(hcsp->HCS_Base + ORC_HDATA, ORC_CMD_GET_NVM);
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    ORC_WR(hcsp->HCS_Base + ORC_HDATA, address);	/* Write address */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    if (waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set   */
	return (FALSE);
    *pDataIn = ORC_RD(hcsp->HCS_Base, ORC_HDATA);
    ORC_WR(hcsp->HCS_Base + ORC_HSTUS, bData);	/* Clear HDI    */

    return (TRUE);
}

/*====================================================================*/
static int
se2_rd_all(ORC_HCS * hcsp)
/*
[Name]

[Description]
 Read SCSI H/A configuration parameters from serial EEPROM
*/
{
    int eep_addr;
    UCHAR *np, chksum;

    for (np = (UCHAR *)nvramp, eep_addr = 0; eep_addr < NVRAM_SIZE;
		 np++, eep_addr++) {
	if (get_NVRAM(hcsp, (unsigned char) eep_addr, np) == FALSE)
		return -1;
//      *np++ = get_NVRAM(hcsp, (unsigned char ) eep_addr);
    }

/*------ Is ckecksum ok ? ------*/
    for (np = (UCHAR *) nvramp + NVRAM_SET, chksum = 0;
		 np < (UCHAR *) nvramp + NVRAM_SIZE - 1; np++) {
	chksum += *np;
    }

    if (nvramp->CheckSum != (UCHAR) chksum) {
	if (dbgf == 3)
	    printf("Check Sum error %02X : %02X\n", nvramp->CheckSum, chksum);
	return -1;
    }
    return 1;
}

/************************************************************************
 Update SCSI H/A configuration parameters from serial EEPROM
*************************************************************************/
void se2_update_all(ORC_HCS * hcsp)
{				/* setup default pattern  */
    int eep_addr;
    UCHAR *np, *np1, chksum = 0;

	/* Calculate checksum first   */
    for (eep_addr = NVRAM_SET, np = (UCHAR *) dftNvRam;
		eep_addr < NVRAM_SIZE - 1; eep_addr++)
	chksum += *np++;
    *np = chksum;

    np = (UCHAR *) dftNvRam;
    np1 = (UCHAR *) nvramp;
    for (eep_addr = 0xC0; eep_addr < NVRAM_SIZE; eep_addr++, np++, np1++) {
	if (*np != *np1) {
		set_NVRAM(hcsp, (unsigned char) eep_addr, *np);
	}
    }
    return;
}

/*************************************************************************
 Function name  : read_eeprom
**************************************************************************/
void read_eeprom(ORC_HCS * hcsp)
{
    if (se2_rd_all(hcsp) != 1) {
	se2_update_all(hcsp);	/* setup default pattern        */
	se2_rd_all(hcsp);	/* load again                   */
    }
}

/*====================================================================*/
static int
OrcWriteEEPCmdReg(ORC_HCS * hcsp, uchar cmd_reg)
/*
 Functio Name	:OrcWriteEEPCmdReg
 Argument	:ORC_HCS *hcsp;	Inic-850 io base address;
		 uchar cmd_reg;	Inic-850 command byte;
 Result		 0:success   FALSE:write-command error
 Description	:Write comannd to EEPROM Cmd Reg.
*/
{
    ORC_WR(hcsp->HCS_Base + ORC_HDATA, ORC_CMD_SET_NVM);  /* Write command */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off */
	return (FALSE);
    return 0;
}

/*====================================================================*/
static void
MakeDiscSyncBit(ORC_HCS *hcsp,
		uchar *pdisc_new, uchar *pstdr_new, uchar *pwide_new,
		uchar *pdisc_en, uchar *pstdr_en, uchar *pwide_en)
/*
 Functio Name	:MakeDiscSyncBit
 Argument	:uchar *pdisc_new;	New value of disconnect enable
					0xFF:no change
					0:disable  1:enable
		 uchar *pstdr_new;	New value of Sync enable
					0:disable  1:enable
		 uchar *pwide_new;	New value of Wide nego enable
					0:disable  1:enable
		 uchar *pdisc_en;	EEP value of disconnect enable
					0:disable  1:enable
		 uchar *pstdr_en;	EEP value of Sync/Async
					0: Sync disable
					1: enable
		 uchar *pwide_en;	EEP value of Wide nego
					0: Wide nego disable
					1: enable
*/
{
    int scsi_id;
    uchar target_val;

    for (scsi_id = 0; scsi_id < 16; scsi_id++) {
        if (dbgf == 2) printf("pdisc_new(%02X) ptdr_new(%02X) pwide_new(%02X)",
		pdisc_new[scsi_id], pstdr_new[scsi_id], pwide_new[scsi_id]);

	target_val = hcsp->TargetFlag[scsi_id] & 0x88; // Bit7,3

	if (pdisc_new[scsi_id] == 0xFF) // no change
	    target_val = target_val | (hcsp->TargetFlag[scsi_id] & 0x40);
	else if (pdisc_new[scsi_id] == 1)
	    target_val = target_val | 0x40;

	if (pstdr_new[scsi_id] == 0xFF) // no change
	    target_val = target_val | (hcsp->TargetFlag[scsi_id] & 0x03);
	else if (pstdr_new[scsi_id] == 0x0F)
	    target_val = target_val | 0x10;
	else
	    target_val = target_val | pstdr_new[scsi_id];

	if (pcb_ver == 0x2400) // PCI31L ?
	    pwide_new[scsi_id] = 0;	// Wide nego disable
	if (pwide_new[scsi_id] == 0xFF) // no change
	    target_val = target_val | (hcsp->TargetFlag[scsi_id] & 0x20);
	else if (pwide_new[scsi_id] == 0)
	    target_val = target_val | 0x20;

	hcsp->TargetFlag[scsi_id] = target_val;
        if (dbgf == 2) printf(" : TargetFlag(%02X)\n", hcsp->TargetFlag[scsi_id]);
    }
}
/*====================================================================*/
static int
WriteDiscSyncBit(ORC_HCS *hcsp,
		uchar *pmode, uchar *pdisc, uchar *pwide,
		uchar *psync, uchar *pdisc_en, uchar *pstdr_en, uchar *pwide_en)
/*
Function Name	:WriteDiscSyncBit
Argument	:uchar *pmode;		20, 10, 5, 0(=Async)
		uchar *pdisc;		0: Disconnect disable
					1: enable
		uchar *pwide;		0: Wide nego disable
					1: enable
	From EEP data
	uchar *psync;		priod(0:20M 2:10M 6:5M) + offset
	uchar *pdisc_en;	0: Disconnect disable 
				1: enable
	uchar *pstdr_en;	0: Sync disable
				1: enable
	uchar *pwide_en;	0: Wide nego disable
				1: enable
=====================================================================*/
{
    int eep_addr;
    UCHAR eep_data, status;
    UCHAR *np, chksum;

    uchar sync_new[16], stdr_new[16];
    int scsi_id, num_targ;

    if (dbgf == 3) {
	for (scsi_id = 0; scsi_id < 16; scsi_id++)
	    printf("Sync(%02X) : Disc(%02X) : Wide(%02X)\n",
			pmode[scsi_id], pdisc[scsi_id], pwide[scsi_id]);
    }

    if (pcbver_set) {
	num_targ = (pcb_ver == 0x2400) ? 8 : 16;
	
	eep_data = pcb_ver & 0x00ff;
	if ((status = set_NVRAM(hcsp, (eep_addr = NVRAM_SET + 0x3C), eep_data))
		!= TRUE) {
	    printf("EEP(%02X) write error : %d\n", eep_addr, status);
	    return status;
	}
	eep_data = pcb_ver / 0x0100;
	if ((status = set_NVRAM(hcsp, (eep_addr = NVRAM_SET + 0x3D), eep_data))
		!= TRUE) {
	    printf("EEP(%02X) write error : %d\n", eep_addr, status);
	    return status;
	}

    	if (dbgf == 3)
		printf("PCB Ver: %04X  # of target : %d\n", pcb_ver, num_targ);

	if ((status = set_NVRAM(hcsp, (eep_addr = NVRAM_SET + 0x14),
		 (eep_data = num_targ))) != TRUE) {
	    printf("EEP(%02X) write error : %d\n", eep_addr, status);
	    return status;
	}
	
    }

    MakeDiscSyncBit(hcsp, pdisc, pmode, pwide, pdisc_en, pstdr_en, pwide_en);
	// Disconnect enable & Sync enable, priod bit
    for (scsi_id = 0, eep_addr = NVRAM_SET + 0x16; scsi_id < 16;
		 scsi_id++, eep_addr++) {
	eep_data = hcsp->TargetFlag[scsi_id];
	if (dbgf == 3) printf("EEP(%02X) : %04X\n", eep_addr, eep_data);
	if ((status = set_NVRAM(hcsp, eep_addr, eep_data)) != TRUE) {
	    printf("EEP(%02X) write error : %d\n", eep_addr, status);
	    return status;
	}
    }

/*------ ckecksum ------*/
    if (se2_rd_all(hcsp) != -1)
	return TRUE;
    for (np = (UCHAR *) nvramp + NVRAM_SET, chksum = 0;
		np < (UCHAR *) nvramp + NVRAM_SIZE - 1; np++) {
	chksum += *np;
//    if (dbgf == 3) printf("(%02X)%02X ", *np, chksum);
    }
    eep_addr = NVRAM_SIZE -1;
    if (dbgf == 3) printf("\nEEP(%04X) check sum : %02X\n", eep_addr, chksum);
    if ((status = set_NVRAM(hcsp, eep_addr, chksum)) != TRUE) {
	printf("EEP(%02X) check write error : %d\n", eep_addr, status);
	    return status;
    }

    return TRUE;
}

/*====================================================================*/
static int
GetSyncDiscFromEEP(ORC_HCS *hcsp,
		uchar *psync,
		uchar *pdisc_en, uchar *pstdr_en, uchar *pwide_en,
		ushort *pPcbVer, ushort *pNumTarg)
/*
 Function name	:GetSyncDiscFromEEP
 Description	:
 Argument	:ORC_HCS *hcsp;
			Inic-850 Adapter info.
		uchar *psync;		priod
				0:40M 1:20M 2:13.3M 3:10.0M 4:8.0M
				5:6.6M 6:5.7M 7:5.0M)
		uchar *pdisc_en;	0: Disconnect disable 
					1: enable
		uchar *pstdr_en;	0: Sync disable
					1: enable
		uchar *pwide_en;	0: Wide disable
					1: enable
		ushort *pPcbVer;	PCB Version
		ushort *pNumTarg;	# of target
*/
{
    ushort eep_addr;
    uchar eep_data, status;

    int scsi_id;

    for (scsi_id = 0, eep_addr = NVRAM_SET + 0x16; scsi_id < 16;
		 scsi_id++, eep_addr++) {
	if ((status = get_NVRAM(hcsp, eep_addr, &eep_data)) != TRUE) {
	    printf("\EEP(%02X) read error : %d\n", eep_addr, status);
	    return status;
	}
	if (dbgf == 4) printf("EEP(%02X) : %02X\n", eep_addr, eep_data);
 	// Target configuration, TCF_EN_TAG
	hcsp->TargetFlag[scsi_id] = eep_data;
	if (eep_data & 0x10) { // bit4:Syncnego
		pstdr_en[scsi_id] = 0; // Async
		psync[scsi_id] = 0x0F; // Async
	} else {
		pstdr_en[scsi_id] = 1; // Syncnego
		psync[scsi_id] = (eep_data & 0x07); // bit2-0:Max.Sync,Speed
	}
	pdisc_en[scsi_id] = (eep_data & 0x40) ? 1 : 0; // bit6:Disconnect
        pwide_en[scsi_id] = (eep_data & 0x20) ? 0 : 1; // bit5:Wide nego
    }

    if ((status = get_NVRAM(hcsp, (eep_addr = NVRAM_SET + 0x3C), &eep_data))
	!= TRUE) {
	    printf("\EEP(%02X) read error : %d\n", eep_addr, status);
	    return status;
    }
    *pPcbVer = eep_data;
    if ((status = get_NVRAM(hcsp, (eep_addr = NVRAM_SET + 0x3D), &eep_data))
	!= TRUE) {
	    printf("\EEP(%02X) read error : %d\n", eep_addr, status);
	    return status;
    }
    *pPcbVer = eep_data * 0x100 + *pPcbVer;

    if ((status = get_NVRAM(hcsp, (eep_addr = NVRAM_SET + 0x14), &eep_data))
	!= TRUE) {
	    printf("\EEP(%02X) read error : %d\n", eep_addr, status);
	    return status;
    }
    *pNumTarg = ((eep_data == 0) ? 16 : eep_data);

    return status;
}

/*====================================================================*/

static void
DispEEP(uchar *psync,
	uchar *pdisc_en, uchar *pstdr_en, uchar *pwide_en,
	ushort PcbVer)
/*
Function name	:DispEEP
Description	:Display EEPROM data (formatted)
Argument	 uchar *psync;
		 uchar *pdisc_en;
		 uchar *pstdr_en;
*/
{
    static char *stdr_txw[] = {	// for Wide Enable
	"       ", " 40.0MB", " 26.6MB", " 20.0MB",
	" 16.0MB", " 13.3MB", " 11.4MB", " 10.0MB"
	};
    static char *stdr_txn[] = { // for Wide Disable
	"       ", " 20.0MB", " 13.3MB", " 10.0MB",
	"  8.0MB", "  6.6MB", "  5.7MB", "  5.0MB"
	};
    int wide_id, scsi_id;

    printf("PCB Version:%04X\n", PcbVer);

    for (wide_id = 0; wide_id < NumTarg; wide_id += 8) {
	printf("\nView EEPROM Setting :");
	for (scsi_id = wide_id; scsi_id < wide_id + 8; scsi_id++)
            printf("  ID-%02d", scsi_id);
	printf("\n");
    
	printf("Enable Disconnect   :");
	for (scsi_id = wide_id; scsi_id < wide_id + 8; scsi_id++) {
	    if (pdisc_en[scsi_id])
		printf("   on  ");
	    else
		printf("  off  ");
	}
	printf("\n");

	printf("Synchronous transfer:");
	for (scsi_id = wide_id; scsi_id < wide_id + 8; scsi_id++) {
	    if (pstdr_en[scsi_id] == 0) {
		printf("  Async");
		continue;
	    }

	    printf("%s", ( (pwide_en[scsi_id])	// wide or narrow ?
				? stdr_txw[psync[scsi_id]]	// wide
				: stdr_txn[psync[scsi_id]]) );	// narrow
	}
	printf("\n");

	printf("Wide Negotiation    :");
	for (scsi_id = wide_id; scsi_id < wide_id + 8; scsi_id++) {
	    if (pwide_en[scsi_id])
		printf("   on  ");
	    else
		printf("  off  ");
	}
	printf("\n");

    }
    printf("\n");

}

/*====================================================================*/
static void
DumpEEP(ORC_HCS *hcsp)
/*
Function name	:DumpEEP
Description	:Dump EEPROM data (hex-mode)
Argument	:ORC_HCS *hcsp;
			Inic-850 Adapter info.
*/
{
    ushort eep_addr, saddr;
    uchar eep_data, status;
    uchar chksum = 0;

    printf("Dump EEP ROM\n");
    printf("  ");
    for (eep_addr = 0; eep_addr < 16; eep_addr++)
	printf(" %02X", eep_addr);
    printf("\n");
    printf("  ");
    for (eep_addr = 0; eep_addr < 16; eep_addr++)
	printf(" --");
    printf("\n");

    saddr = (hex_show == 2) ? 0 : NVRAM_SET;

    for (eep_addr = saddr; eep_addr < NVRAM_SIZE; eep_addr++) {
        if (eep_addr % 32 == 0)
	    printf("%02X", eep_addr);
        else if (eep_addr % 16 == 0)
	    printf("  ");

	if ((status = get_NVRAM(hcsp, eep_addr, &eep_data)) != TRUE)
	    break;
        if (eep_addr != (NVRAM_SIZE - 1) )
	    chksum += eep_data;
	printf(":%02X", eep_data);
	if ((eep_addr + 1) % 16 == 0)
		printf("\n");
    } 
    if (dbgf) printf("\ncheck sum=%02X status = %d\n", chksum, status);
}
/*====================================================================*/
static int
GetHAInfo(ORC_HCS *hcsp)
/*
 Function name	:GetHAInfo
 Argument	:ORC_HCS *hcsp;
*/
{
static struct chip_info {
	ushort id0;
	uchar id1;
	uchar rev;
	ushort status;
	uchar ctrl;
} info;

//    info.id0 = inw(hcsp->HCS_BASE); // ID0
//    info.id1 = inb(io_base+1); // ID1
//    info.rev = inb(io_base+3); // Revision
//    info.status = inw(io_base+0x0e); // Chip Status
//    info.ctrl = inb(io_base+0x0f); // Chip Control

    if (dbgf) {
	printf("ChipID = %04X-%02X / Revision = %02X\n",
			info.id0, info.id1, info.rev);
	printf("Chip Status = %04X, Chip Control = %02X\n",
			info.status, info.ctrl);
    }
//    if (info.id0 == 0xFFFF)
//	return -1; // Not configuraion
    return 0;
}

/*====================================================================*/

typedef struct {
    u_short	vendor, device;
    char	*tag;
    char	*name;
} pci_id_t;

pci_id_t pci_id[] = {
    { 0x1195, 0x0851, "Initio", "INIC-850 on RATOC CB32" },
    { 0x1195, 0x0852, "Initio", "INIC-850 on RATOC CB33" },
    { 0x1195, 0x0853, "Initio", "INIC-850 on RATOC PCI31/33" }
};
#define PCI_COUNT (sizeof(pci_id)/sizeof(pci_id_t))

static void
disp_pci_inf(u_int vendor, u_int device,
			 u_char bus, u_char devfn,
			 u_char irq, u_int io)
{
    static int hdr = 0;

    if (hdr++ == 0)
	printf("\nVendor : Device : Bus : Devfn : PCI slot / func");
    printf("\n %04x  :  %04x  ", vendor, device);
    printf(": %02x  :  %02x   ", bus, devfn);
    printf(":      %02x  /  %02x ", PCI_SLOT(devfn), PCI_FUNC(devfn));
    printf(": %02x  : %08x ", irq, io);
}

static int
pci_probe(int verbose)
/*
 Function name	: pci_probe
 Argument	: int verbose;
*/
{
    char s[256], *t, *name = NULL;
    u_int device, vendor, io, i, n;
    u_char bus, devfn, irq;
    FILE *f;
    
    printf(this_program, this_version);

    if ((f = fopen("/proc/bus/pci/devices", "r")) != NULL) {
	while (fgets(s, 256, f) != NULL) {

	    n = strtoul(s, NULL, 16);
	    bus = (n >> 8); devfn = (n & 0x00ff);
	    n = strtoul(s+5, NULL, 16);
	    vendor = (n >> 16); device = (n & 0xffff);
	    irq = strtoul(s+13, NULL, 16);
	    n = strtoul(s+16, NULL, 16);
	    io = n & 0xfffffff0;
		// pcibios_read_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, &io);
		// pcibios_read_config_byte(bus, devfn, PCI_INTERRUPT_LINE, &irq);

	    if (dbgf)
		disp_pci_inf(vendor, device, bus, devfn, irq, io);

	    for (i = 0; i < PCI_COUNT; i++)
		if ((vendor == pci_id[i].vendor) &&
		    (device == pci_id[i].device)) break;
	    if (i < PCI_COUNT) {
		name = pci_id[i].name;
		// Set device-id bit0-7 for PCI or CB
		dftNvRam[2] = dftNvRam[7] = dftNvRam[0xC2] = dftNvRam[0xC7]
		= (UCHAR) (pci_id[i].device & 0x00ff);
		break;
	    }
	}
    } else {
	printf("Can't open %s", "/proc/bus/pci/devices");
    }
/* Support OLD Kernel
    } else if ((f = fopen("/proc/pci", "r")) != NULL) {
	while (fgets(s, 256, f) != NULL) {
	    t = strstr(s, "Device id=");
	    if (t) {
		device = strtoul(t+10, NULL, 16);
		t = strstr(s, "Vendor id=");
		vendor = strtoul(t+10, NULL, 16);
		for (i = 0; i < PCI_COUNT; i++)
		    if ((vendor == pci_id[i].vendor) &&
			(device == pci_id[i].device)) break;
	    } else
		for (i = 0; i < PCI_COUNT; i++)
		    if (strstr(s, pci_id[i].tag) != NULL) break;
	    if (i != PCI_COUNT) {
		name = pci_id[i].name;
		// Set device-id bit0-7 for PCI or CB
		dftNvRam[2] = dftNvRam[7] = dftNvRam[0xC2] = dftNv[0xC7]
		= (UCHAR) (pci_id[i].device & 0x00ff);
		break;
	    } else {
		t = strstr(s, "CardBus bridge");
		if (t != NULL) {
		    name = t + 16;
		    t = strchr(s, '(');
		    t[-1] = '\0';
		    break;
		}
	    }
	}
    }
Support OLD Kernel */ 

    if (name) {
	if (dbgf)
	    disp_pci_inf(vendor, device, bus, devfn, irq, io);
	hcs.HCS_Base = io;
	hcs.HCS_Intr = irq;
	printf("\n%s found. ", name);
	if (hcs.HCS_Base == 0) {
	    printf("Host adapter is not configured.\n");
	    return -ENODEV;
	}
	printf("I/O Base = %04X\n", hcs.HCS_Base);
	return 0;
    } else {
	printf("\nNot found host adapter.\n");
	return -ENODEV;
    }
}

static void
usage(char *name)
{
    static char *pusage[] = {
	"\t\t[--disc_all=<on | off>] [--disc[0-15]=<on | off>]\n",
	"\t\t[--sync_all=<sync_value>] [--sync[0-15]=<sync_value>]\n",
	"\t\t[--wide_all=<on | off>] [--wide[0-15]=<on | off>]\n\n",
	"  -v : display EEP data\n",
	"  -x : display EEP by hex-code data\n",
	"  --disc_all   : Change disconnect enable.",
	" all scsi id 0 thru 15.\n",
	"  --disc[0-15] : Change disconnect enable.",
	" one scsi id\n",
	"\t\ton: enable  off: disable\n",
	"  --sync_all   : Change synchronous transfer parameter.",
	" all scsi id 0 thru 15.\n",
	"  --sync[0-15] : Change disconnect enable.",
	" one scsi id\n",
	"\t\tsync_value : 40 | 26.6 | 20 | 16 | 13.3 | 10 | no\n",
	"  --wide_all   : Change wide nego enable.",
	" all scsi id 0 thru 15.\n",
	"  --wide[0-15] : Change wide nego enable.",
	" one scsi id\n",
	"\t\ton: enable  off: disable\n",
	"\n",
	NULL
    };

    char **p;

    fprintf(stderr, "usage: %s [-v] [-x]\n", name);
    for (p = pusage; *p != NULL; *p++)
	fprintf(stderr, *p);
}

/*====================================================================*/

static int
SetupOnOffVal(char *on_off, int scsi_id, uchar *on_off_val)
/*
 Function name	:SetupOnOffVal
 Argument	:char *on_off;
		 int scsi_id;
		 uchar *on_off_val;
====================================================================*/
{
    uchar val;
    int i;

    if (strcasecmp(on_off, "on") == 0)
	val = 1;
    else if (strcasecmp(on_off, "off") == 0)
	val = 0;
    else
	return -1;

    if (scsi_id == -1)
	for (i = 0; i < 16; i++)
	    on_off_val[i] = val;
    else
	on_off_val[scsi_id] = val;

    return 0;

}
/*====================================================================*/

static int
SetupSyncVal(char *mode, int scsi_id, uchar *sync_val)
/*
 Function name	:SetupSyncVal
 Argument	:char *mode;
		 int scsi_id;
		 uchar *sync_val;
*/
{
    uchar val;
    int i;

    if (strcmp(optarg, "40") == 0)
	val = 1;
    else if (strcmp(optarg, "40.0") == 0)
 	val = 1;
    else if (strcmp(optarg, "26.6") == 0)
 	val = 2;
    else if (strcmp(optarg, "20") == 0)
 	val = 3;
    else if (strcmp(optarg, "20.0") == 0)
 	val = 3;
    else if (strcmp(optarg, "16") == 0)
 	val = 4;
    else if (strcmp(optarg, "16.0") == 0)
 	val = 4;
    else if (strcmp(optarg, "13.3") == 0)
 	val = 5;
    else if (strcmp(optarg, "11.4") == 0)
 	val = 6;
    else if (strcmp(optarg, "10") == 0)
 	val = 7;
    else if (strcmp(optarg, "10.0") == 0)
 	val = 7;
    else if (strcasecmp(optarg, "no") == 0)
 	val = 0x0f;
    else if (strcasecmp(optarg, "async") == 0)
 	val = 0x0f;
    else
	return -1;

    if (scsi_id == -1)
	for (i = 0; i < 16; i++)
	    sync_val[i] = val;
    else
	sync_val[scsi_id] = val;

    return 0;

}
/*====================================================================*/

static int
SetupPcbVer(char *pcb, ushort *pPcbVer)
/*
 Function name	:SetupPcbVer
 Argument	:char *pcb;
		 ushort *pPcbVer;
*/
{
    int i;
    sscanf(pcb,"%x", &i);
    if (dbgf)
	printf("\npcbver = %04X\n", i);
    if ((i != 0x2300) && (i != 0x2400))
	return -1;

    *pPcbVer = i;

    return 0;
}

/*====================================================================*/

static struct option long_options[] = {
    { "disc_all", 1, 0, 0 },
    { "disc0", 1, 0, 0 }, { "disc1", 1, 0, 0 },
    { "disc2", 1, 0, 0 }, { "disc3", 1, 0, 0 },
    { "disc4", 1, 0, 0 }, { "disc5", 1, 0, 0 },
    { "disc6", 1, 0, 0 }, { "disc7", 1, 0, 0 },
    { "disc8", 1, 0, 0 }, { "disc9", 1, 0, 0 },
    { "disc10", 1, 0, 0 }, { "disc11", 1, 0, 0 },
    { "disc12", 1, 0, 0 }, { "disc13", 1, 0, 0 },
    { "disc14", 1, 0, 0 }, { "disc15", 1, 0, 0 },
    { "sync_all", 1, 0, 0 },
    { "sync0", 1, 0, 0 }, { "sync1", 1, 0, 0 },
    { "sync2", 1, 0, 0 }, { "sync3", 1, 0, 0 },
    { "sync4", 1, 0, 0 }, { "sync5", 1, 0, 0 },
    { "sync6", 1, 0, 0 }, { "sync7", 1, 0, 0 },
    { "sync8", 1, 0, 0 }, { "sync9", 1, 0, 0 },
    { "sync10", 1, 0, 0 }, { "sync11", 1, 0, 0 },
    { "sync12", 1, 0, 0 }, { "sync13", 1, 0, 0 },
    { "sync14", 1, 0, 0 }, { "sync15", 1, 0, 0 },
    { "wide_all", 1, 0, 0 },
    { "wide0", 1, 0, 0 }, { "wide1", 1, 0, 0 },
    { "wide2", 1, 0, 0 }, { "wide3", 1, 0, 0 },
    { "wide4", 1, 0, 0 }, { "wide5", 1, 0, 0 },
    { "wide6", 1, 0, 0 }, { "wide7", 1, 0, 0 },
    { "wide8", 1, 0, 0 }, { "wide9", 1, 0, 0 },
    { "wide10", 1, 0, 0 }, { "wide11", 1, 0, 0 },
    { "wide12", 1, 0, 0 }, { "wide13", 1, 0, 0 },
    { "wide14", 1, 0, 0 }, { "wide15", 1, 0, 0 },
    { "dump", 0, 0, 0 },
    { "pcbver", 1, 0, 0 },
    { "debug", 2, 0, 0 },
    {  0, 0, 0, 0 }
};

static void
SetOptionLong(int option_index)
{
    extern char *optarg;

    if (dbgf) {
	printf("option %s", long_options[option_index].name);
	if (optarg)
	    printf (" with arg %s", optarg);
	printf("\n");
    }

    switch (option_index) {
    case 0: // disc_all
    case 1: // disc0
    case 2: // disc1
    case 3: // disc2
    case 4: // disc3
    case 5: // disc4
    case 6: // disc5
    case 7: // disc6
    case 8: // disc7
    case 9: // disc8
    case 10: // disc9
    case 11: // disc10
    case 12: // disc11
    case 13: // disc12
    case 14: // disc13
    case 15: // disc14
    case 16: // disc15
	SetupOnOffVal(optarg, option_index - 1, disc_val);
	disc_chg = 1;	// mark setup discconet
	break;
    case 17: // sync_all
    case 18: // sync0
    case 19: // sync1
    case 20: // sync2
    case 21: // sync3
    case 22: // sync4
    case 23: // sync5
    case 24: // sync6
    case 25: // sync7
    case 26: // sync8
    case 27: // sync9
    case 28: // sync10
    case 29: // sync11
    case 30: // sync12
    case 31: // sync13
    case 32: // sync14
    case 33: // sync15
	SetupSyncVal(optarg, option_index - 18, sync_val);
	// Not immplement Checking Transfer Rate on narrow mode
	sync_chg = 1;	// mark setup sync mode
	break;
    case 34: // wide_all
    case 35: // wide0
    case 36: // wide1
    case 37: // wide2
    case 38: // wide3
    case 39: // wide4
    case 40: // wide5
    case 41: // wide6
    case 42: // wide7
    case 43: // wide8
    case 44: // wide9
    case 45: // wide10
    case 46: // wide11
    case 47: // wide12
    case 48: // wide13
    case 49: // wide14
    case 50: // wide15
	SetupOnOffVal(optarg, option_index - 35, wide_val);
	wide_chg = 1;	// mark setup discconet
	break;
    case 51: // dump
	hex_show = 2;
	break;
    case 52: // pcbver
	if (SetupPcbVer(optarg, &pcb_ver)) {
		printf("Error: ");
		printf("option %s", long_options[option_index].name);
		if (optarg)
	    		printf (" with arg %s", optarg);
		printf("\n");
		break;
	}
	pcbver_set = 1;	// set PCB version
	break;
    case 53: // debug
	dbgf = (optarg) ? atoi(optarg) : 1;
	break;
    }
}

static void
errexit_iopl(void)
{
    if (iopl(0))
	perror("iopl");
    exit(EXIT_FAILURE);
}

/*====================================================================*/

int main(int argc, char *argv[])
{
    int optch, errflg;
    extern char *optarg;
    int option_index = 0;

    int usage_show = 0;
    errflg = 0;

    while ((optch = getopt_long(argc, argv, "t:hvx",
				long_options, &option_index)) != -1) {
	switch (optch) {
	case 0:
	    SetOptionLong(option_index);
	    break;
//	case 't':
//	    printf("optarg = %s\n", optarg);
//	    break;
	case 'h':
	    usage_show = 1; break;
	case 'v':
	    verbose = 1; break;
	case 'x':
	    hex_show = 1; break;
	default:
	    errflg = 1; break;
	}
    }
    if (usage_show || errflg || (optind < argc)) {
	usage(argv[0]);
	exit(EXIT_FAILURE);
    }

	// find host adapter
    if (pci_probe(verbose) != 0)
	exit(EXIT_FAILURE);

	// start io access
    if (iopl(3)) {
	perror("iopl");
	exit(EXIT_FAILURE);
    }	

//    if (GetChipInfo(&hcs) != 0) {
//	printf("Driver module does not active.\n");
//	printf("You must connect some scsi device, then reboot system.\n\n");
//	errexit_iopl();
//    }

    if (disc_chg || sync_chg || wide_chg || pcbver_set || verbose) {
	errflg = GetSyncDiscFromEEP(&hcs,
			sync_org, disc_org, stdr_org, wide_org,
			&pcb_org, &NumTarg);
        if (errflg != TRUE) errexit_iopl();
    }

    if (disc_chg || sync_chg || wide_chg || pcbver_set) {
	if (dbgf) {
	    printf("Current value of EEP");
	    DispEEP(sync_org, disc_org, stdr_org, wide_org, pcb_org);
	}
	errflg = WriteDiscSyncBit(&hcs, sync_val, disc_val, wide_val,
				sync_org, disc_org, stdr_org, wide_org);
	if (errflg != TRUE) printf("WriteDiscSyncBit : %d\n", errflg);
	errflg = GetSyncDiscFromEEP(&hcs,
			sync_org, disc_org, stdr_org, wide_org,
			&pcb_org, &NumTarg);
    }

    if (disc_chg || sync_chg || wide_chg || verbose) {
	DispEEP(sync_org, disc_org, stdr_org, wide_org, pcb_org);
    } else if (hex_show) {
	DumpEEP(&hcs);
    }

	// end io access
    if (iopl(0)) {
	perror("iopl");
	exit(EXIT_FAILURE);
    }	

    exit(0);
}
